@@ -3,16 +3,19 @@ import json
 import logging
 import numbers
 import threading
+import time
 import uuid
 from inspect import signature
-from typing import Optional, Any, Callable, Mapping, Union
+from typing import Optional, Any, Callable, Mapping, Union, Iterable
 
 from pika import (BlockingConnection, BasicProperties, ConnectionParameters,
                   PlainCredentials)
 from pika.adapters.blocking_connection import BlockingChannel
+from pika.exceptions import (AMQPChannelError, ChannelClosed,
+                             AMQPConnectionError)
 from pika.exchange_type import ExchangeType
 from pika.frame import Method
-from pika.spec import Basic, Queue
+from pika.spec import Basic
 
 __all__ = ('RabbitApp',)
 log = logging.getLogger('snacks')
@@ -31,9 +34,10 @@ class RabbitApp:
             virtual_host: Optional[str] = None,
             username: Optional[str] = None,
             password: Optional[str] = None,
+            retry_timeout: Optional[float] = None
     ) -> None:
         self.host = host or '127.0.0.1'
-        self.port = port or '5762'
+        self.port = port or 5762
         self.default_exchange = default_exchange or ''
         self.virtual_host = virtual_host or '/'
         log.info('RabbitApp configured.')
@@ -48,14 +52,15 @@ class RabbitApp:
         self.params = ConnectionParameters(
             host=host,
             port=int(port),
-            virtual_host=virtual_host,
+            virtual_host=self.virtual_host,
             credentials=self.credentials
         )
-        self.connection = BlockingConnection(self.params)
+        self.connection: BlockingConnection = BlockingConnection(self.params)
         self.channel: BlockingChannel = self.connection.channel()
+        self.retry_timeout: float = retry_timeout or 10.0
 
     @staticmethod
-    def from_dict(dictionary: dict[str, any]) -> 'RabbitApp':
+    def from_dict(dictionary: dict[str, Any]) -> 'RabbitApp':
         return RabbitApp(
             dictionary.get('host'),
             dictionary.get('port'),
@@ -99,6 +104,7 @@ class RabbitApp:
             exchange.
         :return: Method frame from the Exchange.Declare-ok response.
         """
+        self._reset_if_lost()
         return self.channel.exchange_declare(
             exchange or self.default_exchange,
             exchange_type,
@@ -111,12 +117,12 @@ class RabbitApp:
 
     def queue_declare(
             self,
-            queue: str,
+            queue: Optional[str] = None,
             passive: bool = False,
             durable: bool = False,
             exclusive: bool = False,
             auto_delete: bool = False,
-            arguments: Optional[Queue.DeclareOk] = None,
+            arguments: Optional[Mapping[str, Any]] = None,
     ) -> Method:
         """Wrapper for pika.BlockingChannel queue_declare method.
 
@@ -139,8 +145,9 @@ class RabbitApp:
         :param arguments: Custom key/value arguments for the queue.
         :return: Method frame from the Queue.Declare-ok response
         """
+        self._reset_if_lost()
         return self.channel.queue_declare(
-            queue,
+            queue or '',
             passive,
             durable,
             exclusive,
@@ -151,21 +158,22 @@ class RabbitApp:
     def queue_bind(
             self,
             queue: str,
-            exchange: Optional[str] = None,
             routing_key: Optional[str] = None,
+            exchange: Optional[str] = None,
             arguments: Optional[Mapping[str, Any]] = None
     ) -> Method:
-        """ Wrapper for pika.BlockingChannel queue_bind method.
+        """Wrapper for pika.BlockingChannel queue_bind method.
 
         Bind the queue to the specified exchange.
 
         :param queue: The queue to bind to the exchange.
-        :param exchange: The source exchange to bind to.
         :param routing_key: The routing key to bind on.
+        :param exchange: The source exchange to bind to.
         :param arguments: Custom key/value pair arguments for the
             binding.
         :return: Method frame from the Queue.Bind-ok response.
         """
+        self._reset_if_lost()
         return self.channel.queue_bind(
             queue,
             exchange or self.default_exchange,
@@ -173,6 +181,24 @@ class RabbitApp:
             arguments
         )
 
+    def queue_delete(
+            self,
+            queue: str,
+            if_unused: bool = False,
+            if_empty: bool = False
+    ) -> Method:
+        """Wrapper for pika.BlockingChannel queue_delete method.
+
+        Delete a queue from the broker.
+
+        :param queue: The queue to delete.
+        :param if_unused: Only delete if it's unused.
+        :param if_empty: Only delete if the queue is empty.
+        :return: Method frame from the Queue.Delete-ok response.
+        """
+        self._reset_if_lost()
+        return self.channel.queue_delete(queue, if_unused, if_empty)
+
     def publish(
             self,
             body: Any,
@@ -187,6 +213,7 @@ class RabbitApp:
         :param exchange: Exchange to publish to.
         :param kwargs: Keyword args to pass to pika publish.
         """
+        self._reset_if_lost()
         log.debug('Publishing key=%s msg=%s', routing_key, body)
         self.channel.basic_publish(
             exchange=exchange or self.default_exchange,
@@ -213,6 +240,7 @@ class RabbitApp:
         :param time_limit: Number of seconds to wait for a response.
         :param headers: Headers to publish with.
         """
+        self._reset_if_lost()
         log.debug('Publishing key=%s msg=%s', routing_key, body)
         response: Optional[bytes] = None
         corr_id = str(uuid.uuid4())
@@ -225,7 +253,7 @@ class RabbitApp:
         ) -> None:
             nonlocal response
             if props.correlation_id == corr_id:
-                log.debug('RPC response from %s is %s', routing_key, resp_body)
+                log.debug('Response from [%s] is [%s]', routing_key, resp_body)
                 response = resp_body
 
         result = self.channel.queue_declare(queue='', exclusive=True)
@@ -250,39 +278,131 @@ class RabbitApp:
 
     def listener(
             self,
-            queues: list[str],
-            serialize=None
+            routing_keys: Union[list[str], str],
+            exchange: Optional[str] = None,
+            serialize: Optional[Callable] = None,
+            declare_arguments: Optional[Mapping[str, Any]] = None,
+            bind_arguments: Optional[Mapping[str, Any]] = None
     ) -> Callable:
-        """Decorate a function to make it a rabbit consumer.
+        """Decorate a callable to generate queues and consume from them.
 
-        The decorated function can have as parameters, the message body,
-        the method, the properties, or no parameters.
+        A new non-durable, auto-deleting queue will be generated for
+        each provided routing key.
+        The decorated function can have as parameters, any or all of,
+        message body, method, and properties.
 
         A new channel and thread is created for each consumer.
 
-        :param queues: The queues to consume from.
+        :param routing_keys: Key or keys to generate queues for.
+        :param exchange: The queues to consume from.
+        :param serialize: Callable to serialize message bodies.
+        :param declare_arguments: Arguments for queue_declare.
+        :param bind_arguments: Arguments for queue_bind.
+        :return: Function decorated to be a rabbit consumer.
+        """
+        self._reset_if_lost()
+        rks = [routing_keys] if isinstance(routing_keys, str) else routing_keys
+
+        # If a connection is lost, the auto-deleting queues will all be
+        # gone so reconnecting won't work. We pass the consume logic
+        # this queue_factory to generate the initially and whenever a
+        # connection fails.
+        def queue_factory() -> list[str]:
+            queues = []
+            for key in rks:
+                q = self.queue_declare(
+                    arguments=declare_arguments,
+                    auto_delete=True
+                ).method.queue
+                self.queue_bind(q, key, exchange, arguments=bind_arguments)
+                queues.append(q)
+            return queues
+
+        def wrapper(fun: Callable) -> Any:
+            thread = threading.Thread(
+                target=self._consume,
+                args=(fun, [], serialize, queue_factory)
+            )
+            thread.daemon = True
+            thread.start()
+            for rk in rks:
+                log.info('[%s] consuming queue for key [%s]', fun.__name__, rk)
+
+        return wrapper
+
+    def consumer(
+            self,
+            queues: Union[list[str], str],
+            serialize: Optional[Callable] = None
+    ) -> Callable:
+        """Decorate a callable to consume from one or more queues.
+
+        The decorated function can have as parameters, any or all of,
+        message body, method, and properties.
+
+        A new channel and thread is created for each consumer.
+
+        :param queues: The queue or queues to consume from.
         :param serialize: Callable to serialize message bodies.
         :return: Function decorated to be a rabbit consumer.
         """
-        # Threads cannot share a connection or channel.
-        channel = BlockingConnection(self.params).channel()
 
-        def wrapper(fun: Callable) -> None:
-            original_name = fun.__name__
-            for queue in queues:
-                fun.__name__ = f'listener_{queue}'
-                _listen(channel, queue, fun, serialize)
-                log.info(
-                    'Function %s subscribed to queue %s.',
-                    original_name,
-                    queue
-                )
-            thread = threading.Thread(target=_consume, args=(channel,))
+        def wrapper(fun: Callable) -> Any:
+            thread = threading.Thread(
+                target=self._consume,
+                args=(fun, queues, serialize)
+            )
             thread.daemon = True
             thread.start()
 
         return wrapper
 
+    def _consume(
+            self,
+            fun: Callable,
+            queues: Union[Iterable[str], str],
+            serialize: Optional[Callable] = None,
+            queue_factory: Optional[Callable[[], list[str]]] = None
+    ) -> None:
+        running = True
+        channel = None
+        while running:
+            try:
+                if self.connection.is_closed:
+                    self._reset_if_lost()
+                if not channel or channel.is_closed:
+                    # New channel since threads can't share channel.
+                    channel = self.connection.channel()
+
+                def consume(q: str) -> None:
+                    _listen(channel, q, fun, serialize)
+                    level = log.debug if q.startswith('amq.gen') else log.info
+                    level('[%s] consuming queue [%s]', fun.__name__, q)
+
+                if queue_factory:
+                    queues = queue_factory()
+                queues = [queues] if isinstance(queues, str) else queues
+                [consume(queue) for queue in queues]
+                channel.start_consuming()
+            except (ChannelClosed, AMQPChannelError, AMQPConnectionError) as e:
+                log.error('%s on channel for %s', type(e).__name__, queues)
+                log.info(
+                    'Retrying connection in [%s] seconds...',
+                    self.retry_timeout
+                )
+            except KeyboardInterrupt:
+                running = False
+                if channel:
+                    channel.stop_consuming()
+                    channel.close()
+            time.sleep(self.retry_timeout)
+
+    def _reset_if_lost(self) -> None:
+        if self.connection.is_closed:
+            self.connection = BlockingConnection(self.params)
+        if self.channel.is_closed:
+            self.channel = self.connection.channel()
+
 
 def _listen(
         channel: BlockingChannel,
@@ -298,9 +418,14 @@ def _listen(
             properties: BasicProperties,
             body: bytes
     ) -> None:
-        log.debug('Queue %s received %s.', queue, body)
+        log.debug('Queue [%s] received [%s]', queue, body)
         kwargs: dict[str, Any] = {}
         for name, param in sig.parameters.items():
+            # FIXME Typing annotations break this.
+            #  If type is Optional, get rid of the Optional.
+            #  If type is Union try mapping to all, take what works.
+            #  If type is Iterable default to list.
+            #  If type is Mapping default to dict.
             if name == 'self':
                 kwargs[name] = None
             elif param.annotation == BlockingChannel:
@@ -338,7 +463,7 @@ def _listen(
                 )
                 ch.basic_ack(delivery_tag=method.delivery_tag)
         except Exception as e:
-            log.exception(msg=f'{type(e).__name__}: {e}', exc_info=e)
+            log.exception(msg=f'{type(e).__name__}:{e}', exc_info=e)
 
     channel.basic_qos(prefetch_count=1)
     channel.basic_consume(
@@ -348,14 +473,6 @@ def _listen(
     )
 
 
-def _consume(channel: BlockingChannel) -> None:
-    try:
-        channel.start_consuming()
-    except KeyboardInterrupt:
-        channel.stop_consuming()
-        channel.close()
-
-
 def _def_serialize(body: Any) -> bytes:
     if 'to_json' in dir(body):
         return body.to_json().encode()
